/*
 * Copyright (C) 2010 Sven Killig
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package de.killig.gamepadime;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;

import android.inputmethodservice.InputMethodService;
import android.os.Handler;
import android.util.Log;
import android.view.KeyEvent;
import android.widget.Toast;

import static android.view.KeyEvent.*; 

public class GamepadIMEThread extends Thread {
  private static final String LOG_TAG = "GamepadIMEThread";
  private final static short JS_EVENT_BUTTON=0x01;
  private final static short JS_EVENT_AXIS  =0x02;
  //private final static short JS_EVENT_INIT  =0x80;
	private File dev;
	private static /*InputConnection*/ InputMethodService i;
	private static Handler mHandler = new Handler();
	
	public GamepadIMEThread(/*InputConnection*/ InputMethodService i, File dev) {
		this.dev=dev;
		GamepadIMEThread.i=i;
	}
	
	private /*static*/ void sendKeyEvent(int action, int code) {
		Log.d(LOG_TAG, dev.getName()+": action="+action+"; code="+code);
		i.getCurrentInputConnection().sendKeyEvent(new KeyEvent(action, code));
	}

  static void handleException(final Exception e) {
		e.printStackTrace();
		mHandler.post(
        new Runnable() { 
            public void run() {
              Toast.makeText(i, e.getMessage(), Toast.LENGTH_LONG).show();
            }
        }
    );
  }

	public void run() {
		InputStream is=null;
  	byte[] buffer=new byte[8];
		ByteArrayInputStream bais=new ByteArrayInputStream(buffer);
		// 4,25 KB/js: 
    int[] button_actions    =new int[256];
  	byte[] axis_hold_flags  =new byte[256];
  	int[][] axis_actions    =new int[256][2];
  	short[][] axis_threshold=new short[256][2];
  	/*
  	 * 1:x:2 y:3 b:2
  	 * 2:x:0 y:1 b:0
  	*/
  	/*button_actions[0]=KeyEvent.KEYCODE_DPAD_DOWN;
  	button_actions[1]=KeyEvent.KEYCODE_DPAD_LEFT;
  	button_actions[2]=KeyEvent.KEYCODE_DPAD_RIGHT;
  	button_actions[3]=KeyEvent.KEYCODE_DPAD_UP;*/
  	button_actions[ 0]=KEYCODE_P;
  	axis_actions[0][0]=KEYCODE_A;
  	axis_actions[0][1]=KEYCODE_D;
  	axis_actions[1][0]=KEYCODE_W;
  	axis_actions[1][1]=KEYCODE_S;
  	
  	/*for(int i=1; i<12; i++) {
    	button_actions[i]=KeyEvent.KEYCODE_DPAD_CENTER;
  	}*/
  	button_actions[ 1]=KEYCODE_DPAD_CENTER;
  	button_actions[ 2]=KEYCODE_0;
  	button_actions[ 3]=KEYCODE_1;
  	button_actions[ 4]=KEYCODE_2;
  	button_actions[ 5]=KEYCODE_3;
  	button_actions[ 6]=KEYCODE_4;
  	button_actions[ 7]=KEYCODE_5;
  	button_actions[ 8]=KEYCODE_6;
  	button_actions[ 9]=KEYCODE_7;
  	button_actions[10]=KEYCODE_8;
  	button_actions[11]=KEYCODE_9;

  	for(int i=1; i<=2; i++) {
    	axis_actions[i*2  ][0]=KEYCODE_DPAD_LEFT;
    	axis_actions[i*2  ][1]=KEYCODE_DPAD_RIGHT;
    	axis_actions[i*2+1][0]=KEYCODE_DPAD_UP;
    	axis_actions[i*2+1][1]=KEYCODE_DPAD_DOWN;
  	}

  	for(int i=0; i<axis_threshold.length; i++) {
  		axis_threshold[i][0]=-32767/2;
  		axis_threshold[i][1]= 32767/2;
  	}

  	try {
    	is=new FileInputStream(dev);
    	bais.skip(4);
    	bais.mark(0);
			while(!isInterrupted()) {
				is.read(buffer);//),0,8);
				
				/*String s=""; 
				for(int i=0; i<8; i++) s+=buffer[i]+", ";
				Log.d(this.getClass().getSimpleName(), s);*/
				
				/*bais.reset();
				long time=bais.read()+256*bais.read()+256*256*bais.read()+256*256*256*bais.read();*/
				bais.reset();
				short value=(short)(bais.read()+256*bais.read());
				short type=(short)bais.read();
				short number=(short)bais.read();
				Log.v(LOG_TAG, /*"time="+time+*/"; value="+value+"; type="+type+"; number="+number);
				

				switch(type) {
					case JS_EVENT_BUTTON:
						if(button_actions[number]!=0) sendKeyEvent(value!=0 ? ACTION_DOWN : ACTION_UP, button_actions[number]);
						break;
					case JS_EVENT_AXIS:
						if(axis_actions[number][0]!=0 || axis_actions[number][1]!=0) {
							if(axis_hold_flags[number]!=0) {
								switch(axis_hold_flags[number]) {
									case 1:
										if(value>axis_threshold[number][0]) {
											sendKeyEvent(ACTION_UP, axis_actions[number][0]);
											axis_hold_flags[number]=0;
										}
										break;
									case 2:
										if(value<axis_threshold[number][1]) {
											sendKeyEvent(ACTION_UP, axis_actions[number][1]);
											axis_hold_flags[number]=0;
										}
										break;
								}
//								axis_hold_flags[number]=0;
							}
							if(value<axis_threshold[number][0] && axis_hold_flags[number]!=1) { 
								if(axis_actions[number][0]==0) break;
								sendKeyEvent(ACTION_DOWN, axis_actions[number][0]);
								axis_hold_flags[number]=1;
							}
							if(value>axis_threshold[number][1] && axis_hold_flags[number]!=2) {
								if(axis_actions[number][1]==0) break;
								sendKeyEvent(ACTION_DOWN, axis_actions[number][1]);
								axis_hold_flags[number]=2;
							}		
						}	    
						break;
				}
			
			
			}
		} catch (Exception e) {
			handleException(e);
		} finally {
			try {
				if(is!=null) is.close();
			} catch (Exception e) {
				handleException(e);
			}
		}
  }

}
